package de.uni_passau.fim.seibt.gitwrapper.repo;

import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;

import java.util.*;
import java.util.function.Function;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * A {@link Commit} made in a {@link Repository}.
 */
public class Commit {

    private static final Logger LOG = Logger.getLogger(Commit.class.getCanonicalName());

    private GitWrapper git;
    private Repository repo;

    private String id;

    /**
     * Constructs a new {@link Commit} with the given <code>id</code> made in the <code>repo</code>.
     *
     * @param repo
     *         the {@link Repository} the {@link Commit} was made in
     * @param id
     *         the ID of the {@link Commit}
     */
    Commit(Repository repo, String id) {
        this.git = repo.getGit();
        this.repo = repo;
        this.id = id;
    }

    /**
     * Returns the parents of this {@link Commit}.
     *
     * @return the parent {@link Commit} objects
     */
    public List<Commit> getParents() {
        Optional<ExecRes> revList = git.exec(repo.getDir(), "rev-list", "--parents", "-n", "1", id);
        Function<ExecRes, List<Commit>> toParentsList = res -> {
            String[] ids = res.stdOut.split("\\s+");

            LOG.fine(() -> String.format("Found %d parents for %s.", ids.length - 1, this));
            LOG.finer(() -> String.format("Commit id and parents are:%n%s", String.join(System.lineSeparator(), ids)));

            return Arrays.stream(ids).skip(1).map(repo::getCommitUnchecked).collect(Collectors.toList());
        };

        return revList.map(toParentsList).orElse(Collections.emptyList());
    }

    /**
     * Optionally returns the merge base for <code>this</code> and <code>other</code>. The <code>other</code> commit
     * must be part of the same {@link Repository} this {@link Commit} is.
     *
     * @param other
     *         the other {@link Commit}
     * @return the merge base or an empty {@link Optional} if there is no merge base or an exception occurred
     */
    public Optional<Commit> getMergeBase(Commit other) {

        if (!repo.equals(other.repo)) {
            LOG.warning(() -> {
                String msg = "Failed to obtain a merge base for %s and %s as they are not from the same repository.";
                return String.format(msg, this, other);
            });

            return Optional.empty();
        }

        Optional<ExecRes> mergeBase = git.exec(repo.getDir(), "merge-base", getId(), other.getId());
        Function<ExecRes, Commit> toCommit = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to obtain a merge base for %s and %s.", this, other));
                return null;
            }

            Commit base = repo.getCommitUnchecked(res.stdOut.trim());

            LOG.fine(() -> String.format("Commits %s and %s have the merge base %s.", this, other, base));

            return base;
        };

        return mergeBase.map(toCommit);
    }

    /**
     * Returns the ID of this commit.
     *
     * @return the ID
     */
    public String getId() {
        return id;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Commit commit = (Commit) o;
        return Objects.equals(repo, commit.repo) && Objects.equals(id, commit.id);
    }

    @Override
    public int hashCode() {
        return Objects.hash(repo, id);
    }

    @Override
    public String toString() {
        return String.valueOf(id);
    }
}
