package de.uni_passau.fim.seibt.gitwrapper.repo;


import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;
import de.uni_passau.fim.seibt.gitwrapper.process.ToolNotWorkingException;
import de.uni_passau.fim.seibt.gitwrapper.process.ToolWrapper;

import java.io.File;
import java.util.Optional;
import java.util.Scanner;
import java.util.function.Function;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.asList;

/**
 * A wrapper for executing git commands.
 */
public class GitWrapper extends ToolWrapper {

    public static final int EXIT_SUCCESS = 0;
    public static final int EXIT_FAIL = 1;

    private static final Logger LOG = Logger.getLogger(GitWrapper.class.getCanonicalName());

    private static final String FATAL_PREFIX = "fatal";
    private static final String ERROR_PREFIX = "error";

    private static final Pattern CLONING_INTO = Pattern.compile("Cloning into '(.*)'\\.\\.\\.");
    private static final Pattern ALREADY_EXISTS = Pattern.compile("fatal: destination path '(.*)' already exists and is not an empty directory\\.");

    private static final String GIT_VERSION_C = "--version";
    private static final String GIT_VERSION_PREFIX = "git version";

    /**
     * Constructs a new {@link GitWrapper}
     *
     * @param git the git command, e.g. '/usr/bin/git' or just 'git'
     */
    public GitWrapper(String git) throws ToolNotWorkingException {
        super(git);
    }

    @Override
    protected boolean isWorking() {
        Function<ExecRes, Boolean> checkPrefix = execRes -> execRes.stdOut.startsWith(GIT_VERSION_PREFIX);
        return exec(new File("."), true, asList(GIT_VERSION_C)).map(checkPrefix).orElse(false);
    }

    /**
     * Attempts to clone the git repository at the given <code>url</code> into a directory under <code>parentDir</code>.
     * If <code>parentDir</code> already contains a git repository with the name that would be produced when cloning
     * <code>url</code> it is assumed to be a clone of <code>url</code> and returned as a {@link Repository}.
     *
     * @param parentDir
     *         the directory in which to place the clone of <code>url</code>
     * @param url
     *         the git repository to clone
     * @param fetch
     *         whether to perform a fetch of the {@link Repository} if a clone of it was already found on disk
     * @return the resulting {@link Repository} or an empty {@link Optional} if there is an error
     */
    public Optional<Repository> clone(File parentDir, String url, boolean fetch) {
        LOG.fine(() -> String.format("Cloning '%s'.", url));

        Optional<ExecRes> res = exec(parentDir, "clone", url);
        Function<ExecRes, Repository> toRepo = execRes -> {
            Scanner sc = new Scanner(execRes.stdOut);

            if (!sc.hasNextLine()) {
                LOG.warning(() -> String.format("No output while cloning '%s'.", url));
                return null;
            }

            String firstLine = sc.nextLine();
            Matcher matcher;

            if (execRes.exitCode == EXIT_SUCCESS) {
                matcher = CLONING_INTO.matcher(firstLine);
            } else {
                matcher = ALREADY_EXISTS.matcher(firstLine);
            }

            if (!matcher.find()) {
                LOG.warning(() -> String.format("Unexpected output while cloning '%s'.", url));
                return null;
            }

            String name = matcher.group(1);
            File repoDir = new File(parentDir, name);
            String repoPath = repoDir.getAbsolutePath();

            if (execRes.exitCode == EXIT_SUCCESS) {
                LOG.fine(() -> String.format("Cloned '%s' into '%s'.", url, repoPath));
                return new Repository(this, url, repoDir);
            } else {
                LOG.fine(() -> String.format("Found '%s' which would be produced for '%s'", repoPath, url));

                if (!isGitDir(repoDir)) {
                    LOG.warning(() -> String.format("'%s' is not a git repository.", repoPath));
                    return null;
                }

                LOG.fine(() -> String.format("'%s' is a git repository. Assuming it is a clone of '%s'.", repoPath, url));
                Repository repo = new Repository(this, url, repoDir);

                if (fetch && !repo.fetch()) {
                    LOG.warning("Could not fetch " + repo + ". It may be out of date.");
                }

                return repo;
            }
        };

        return res.map(toRepo);
    }

    /**
     * Returns whether the given directory (or any of its parent directories) is a git repository. Returns false if
     * <code>directory</code> is not a directory.
     *
     * @param directory
     *         the directory to check
     * @return true iff the given directory (or any of its parent directories) is a git repository
     */
    public boolean isGitDir(File directory) {

        if (!directory.isDirectory()) {
            return false;
        }

        Optional<ExecRes> status = exec(directory, "rev-parse", "--is-inside-git-dir");
        return status.map(res -> !failed(res)).orElse(false);
    }

    /**
     * Returns whether the given git command failed. This relies on exit code first and then on the assumption that
     * the output (in case of failure) starts with either "{@value FATAL_PREFIX}" or "{@value ERROR_PREFIX}".
     *
     * @param res the {@link ExecRes} to check
     * @return whether the git command failed
     */
    public boolean failed(ExecRes res) {
        return res.exitCode != EXIT_SUCCESS || res.stdOut.startsWith(FATAL_PREFIX) || res.stdOut.startsWith(ERROR_PREFIX);
    }

    /**
     * Executes '&lt;git command&gt &lt;parameters&gt' and returns the exit code and the output of the command. If there is an
     * exception executing the command an empty {@link Optional} will be returned.
     *
     * @param workingDirectory
     *         the working directory for the command execution
     * @param parameters
     *         the git parameters
     * @return the result of the execution
     */
    public Optional<ExecRes> exec(File workingDirectory, String... parameters) {
        return exec(workingDirectory, true, asList(parameters));
    }
}
